package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kommons.util.MailboxExecutorService
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.render.LayeredRenderer
import ru.dbotthepony.kstarbound.defs.JsonDriven
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.LightCalculator
import ru.dbotthepony.kstarbound.world.World
import java.util.UUID
import kotlin.concurrent.withLock

abstract class AbstractEntity(path: String) : JsonDriven(path) {
	/**
	 * The chunk this entity resides in
	 */
	var chunk: Chunk<*, *>? = null
		set(value) {
			if (innerWorld == null) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			} else if (value != null && innerWorld != value.world) {
				throw IllegalArgumentException("$this belongs to $innerWorld, $value belongs to ${value.world}")
			} else if (value == field) {
				return
			}

			val oldChunk = field
			field = value

			world.lock.withLock {
				if (oldChunk == null && value != null) {
					world.orphanedEntities.remove(this)
					value.addEntity(this)
				} else if (oldChunk != null && value == null) {
					world.orphanedEntities.add(this)
					oldChunk.removeEntity(this)
				} else if (oldChunk != null && value != null) {
					value.transferEntity(this, oldChunk)
				}
			}
		}

	var uuid: UUID = UUID.randomUUID()
	abstract val chunkPos: ChunkPos

	var mailbox = MailboxExecutorService()
		private set

	private var innerWorld: World<*, *>? = null

	val world: World<*, *>
		get() = innerWorld ?: throw IllegalStateException("Not in world")

	val isSpawned: Boolean
		get() = innerWorld != null

	/**
	 * Whenever this entity should be removed when chunk containing it is being unloaded
	 *
	 * Returning false will also stop entity from being saved to disk, and render entity orphaned
	 * when chunk containing it will get unloaded
	 */
	open val isApplicableForUnloading: Boolean
		get() = true

	protected open fun onSpawn(world: World<*, *>) { }
	protected open fun onRemove(world: World<*, *>) { }

	/**
	 * MUST be called by [World] itself
	 */
	fun spawn(world: World<*, *>) {
		if (innerWorld != null)
			throw IllegalStateException("Already spawned (in world $innerWorld)")

		world.ensureSameThread()

		if (mailbox.isShutdown)
			mailbox = MailboxExecutorService()

		innerWorld = world
		world.entities.add(this)
		world.orphanedEntities.add(this)
		onSpawn(world)
	}

	fun remove() {
		val world = innerWorld ?: throw IllegalStateException("Not in world")
		world.ensureSameThread()

		mailbox.shutdownNow()
		chunk = null
		world.entities.remove(this)
		world.orphanedEntities.remove(this)
		onRemove(world)
		innerWorld = null
	}

	open val isRemote: Boolean
		get() = innerWorld?.isRemote ?: false

	fun think() {
		thinkShared()

		if (isRemote) {
			thinkRemote()
		} else {
			thinkLocal()
		}
	}

	protected open fun thinkShared() {
		mailbox.executeQueuedTasks()
	}

	protected open fun thinkRemote() {

	}

	protected open fun thinkLocal() {

	}

	open fun render(client: StarboundClient, layers: LayeredRenderer) {

	}

	open fun addLights(lightCalculator: LightCalculator, xOffset: Int, yOffset: Int) {

	}
}
